/*
 *  Schedule display v3 
 *  Arduino Nano v3 + OLED display 64x32
 *  Shalex 2017 
 *  http://forum.modelldepo.ru/showthread.php?t=18889
 *  
 *    
 *  CV0 - address of decoder and the first (from six) address for control commands
 *  CV1 - contrast (0..100?)
 *  
 *  to control send a turnout(switch) command with address from CV0 
 *  add to address a number from 1 to 6 to display other schedule lines
 *  
 *  if the Direction = sraight in the command then line adds to Schedule display
 *  and will be removed from display in other case
 *  
 */
 
#include <LocoNet.h>
#include <EEPROMex.h>
#include "U8glib.h"
#include <QueueList.h>
#include <SimpleList.h>

#define DRAW_ON

#ifdef DRAW_ON
U8GLIB_SSD1306_64X48 u8g(U8G_I2C_OPT_NONE);  // I2C / TWI 
#endif

#define HEAD_TEXT "Shalex"
#define LINE_HEIGHT 4
#define LINES_Y_OFFSET 9
#define MAX_LINES 10
#define MAX_LINES_ON_DISPLAY 5

const uint8_t L[MAX_LINES][24] PROGMEM = 
{{                                                                                                                  
40  ,             245 ,             215 ,             165 ,             246 ,             3 ,             109 ,             143 ,
57  ,             119 ,             191 ,             255 ,             222 ,             3 ,             239 ,             143 ,
57  ,             223 ,             254 ,             231 ,             118 ,             3 ,             109 ,             143 
},{                                                                                                                 
89  ,             254 ,             249 ,             254 ,             224 ,             3 ,             109 ,             143 ,
249 ,             159 ,             151 ,             253 ,             224 ,             3 ,             239 ,             143 ,
241 ,             251 ,             249 ,             119 ,             160 ,             3 ,             109 ,             143 
},{                                                                                                                 
88  ,             222 ,             111 ,             220 ,             0 ,             3 ,             109 ,             143 ,
241 ,             237 ,             239 ,             124 ,             0 ,             3 ,             239 ,             143 ,
121 ,             254 ,             125 ,             248 ,             0 ,             3 ,             109 ,             143 
},{                                                                                                                 
88  ,             219 ,             39  ,             223 ,             190 ,             3 ,             109 ,             143 ,
121 ,             95  ,             253 ,             254 ,             255 ,             3 ,             239 ,             143 ,
121 ,             255 ,             63  ,             91  ,             245 ,             3 ,             109 ,             143 
},{                                                                                                                 
120 ,             219 ,             78  ,             181 ,             93  ,             99  ,             109 ,             143 ,
233 ,             95  ,             122 ,             253 ,             255 ,             243 ,             239 ,             143 ,
249 ,             251 ,             207 ,             247 ,             235 ,             179 ,             109 ,             143 
},{                                                                                                                 
233 ,             125 ,             143 ,             93  ,             122 ,             3 ,             109 ,             143 ,
249 ,             247 ,             247 ,             123 ,             254 ,             3 ,             239 ,             143 ,
121 ,             221 ,             157 ,             255 ,             238 ,             3 ,             109 ,             143 
},{                                                                                                                 
169 ,             254 ,             231 ,             251 ,             224 ,             3 ,             109 ,             143 ,
121 ,             253 ,             254 ,             126 ,             64  ,             3 ,             239 ,             143 ,
241 ,             119 ,             167 ,             239 ,             224 ,             3 ,             109 ,             143 
},{                                                                                                                 
89  ,             191 ,             113 ,             188 ,             0 ,             3 ,             109 ,             143 ,
81  ,             189 ,             255 ,             216 ,             0 ,             3 ,             239 ,             143 ,
121 ,             247 ,             227 ,             252 ,             0 ,             3 ,             109 ,             143 
},{                                                                                                                 
81  ,             62  ,             253 ,             240 ,             219 ,             3 ,             109 ,             143 ,
249 ,             239 ,             247 ,             255 ,             95  ,             3 ,             239 ,             143 ,
249 ,             250 ,             223 ,             169 ,             255 ,             3 ,             109 ,             143 
},{                                                                                                                 
113 ,             214 ,             171 ,             172 ,             54  ,             211 ,             109 ,             143 ,
217 ,             95  ,             191 ,             255 ,             215 ,             211 ,             239 ,             143 ,
121 ,             254 ,             253 ,             118 ,             126 ,             243 ,             109 ,             143 
}};                                                                                                                 
                                                                                                             


const uint8_t LINE_EMPTY[24] PROGMEM = {
0  ,             0 ,             0 ,             0 ,             0 ,             0 ,             0 ,             0 ,
0 ,             0 ,             0 ,             0 ,             0 ,             0 ,             0 ,             6 ,
112 ,             255 ,             255 ,             255 ,             255 ,             193 ,             199 ,             0   
};

#define EVENT_TIME_UPDATE 1
#define EVENT_LINE_ADD    2
#define EVENT_LINE_DEL    3


typedef struct  {
  uint8_t EventType;
  uint8_t Arg1;
  uint8_t Arg2;
} TEvent;

volatile uint8_t CurrentHours;
volatile uint8_t CurrentMinutes;

QueueList <TEvent> ScheduleQueue;
SimpleList<uint8_t> Lines;

#define LOCONET_TX_PIN 7

// Item Number (Art.-Nr.): 50010
#define ARTNR 5010
unsigned long EEPROM_VERSION = 01; //0-99
unsigned long EEPROM_KEY = (ARTNR - 5000) * 100 + EEPROM_VERSION;

#define LNCV_COUNT 2

#define CV_ADDR 0
#define CV_CONTRAST 1

volatile uint16_t lncv[LNCV_COUNT] = {12, 50};

LocoNetCVClass lnCV;
LocoNetFastClockClass  FastClock;
unsigned long     LastFastClockTick;

volatile boolean programmingMode = 0;
volatile unsigned long programmingModeStart = 0;
unsigned long ledCycle = 500;
unsigned long PROG_MODE_TIMEOUT = 30000; //mills, 60sec

//for led indication
unsigned long LedLastBlink = 0;
byte          LedState = 0;

const int ALARM_LED = 13;

lnMsg        *LnPacket;

boolean isTime(unsigned long *timeMark, unsigned long timeInterval) {
    unsigned long timeNow = millis();
    if ( timeNow - *timeMark >= timeInterval) {
        *timeMark += timeInterval;
        return true;
    }    
    return false;
}

void drawTime(uint8_t hours, uint8_t mins){
 #ifdef DRAW_ON
  String sT = String(String(hours / 10,DEC) + String(hours % 10,DEC) + ":" + String(mins / 10,DEC) + String(mins % 10,DEC));
  char charBufVar[6];
  sT.toCharArray(charBufVar, sizeof(charBufVar));
  
  u8g.setRGB(255,0,0);
  u8g.setFont(u8g_font_4x6);
  u8g.drawStr(44, 0 + 16, charBufVar);
 #endif
}

void addScheduleEvent(uint8_t EventType, uint8_t Arg1, uint8_t Arg2){
  TEvent Event = {EventType, Arg1, Arg2};
  ScheduleQueue.push(Event);
  Serial.print("Event add:");
  Serial.println(Event.EventType);
  Serial.print(", arg1:");
  Serial.print(Event.Arg1);
  Serial.print(", arg2:");
  Serial.println(Event.Arg2);
}

void drawHeader() {
  #ifdef DRAW_ON
  u8g.setRGB(255,0,0);
  u8g.setFont(u8g_font_4x6);
  u8g.drawStr(0, 0 + 16, HEAD_TEXT);
  #endif
}
void drawScLines() {
  #ifdef DRAW_ON
  u8g.setRGB(255,0,0);
  uint8_t c = 0;
  for (SimpleList<uint8_t>::iterator itr = Lines.begin(); itr != Lines.end();){
      if (c < MAX_LINES_ON_DISPLAY)
        u8g.drawBitmapP(0, LINES_Y_OFFSET + 16 + (c * LINE_HEIGHT), 8, 3, L[(*itr)]);
       ++itr;
       c++; 
  }    
  for (uint8_t i = c; i < MAX_LINES_ON_DISPLAY; i++) 
    u8g.drawBitmapP(0, LINES_Y_OFFSET + 16 + (i * LINE_HEIGHT), 8, 3, LINE_EMPTY);
    
  #endif
}

unsigned long last_t;
bool s;
uint8_t H = 0;
uint8_t M = 0;

void TestIdle(){
    if ((millis() - last_t) > 60000) {
      s = !s;
      M++;
      if (M > 59) {
        if (H > 23) H = 0;
        M = 0;
      }
      addScheduleEvent(EVENT_TIME_UPDATE, H, M);
      last_t = millis();
  }
}

void PictureIdle(void) {
  TEvent Event;
  if (!ScheduleQueue.isEmpty()) {
    
    Event = ScheduleQueue.pop();
    
    Serial.print("Event pop t:");
    Serial.print(Event.EventType);
    Serial.print(", arg1:");
    Serial.print(Event.Arg1);
    Serial.print(", arg2:");
    Serial.println(Event.Arg2);

    bool flag = false;
               
    switch (Event.EventType) {
      case EVENT_TIME_UPDATE:
            CurrentHours = Event.Arg1;
            CurrentMinutes = Event.Arg2;
            break;
      case EVENT_LINE_ADD:
        for (SimpleList<uint8_t>::iterator itr = Lines.begin(); itr != Lines.end();){
            if ((*itr) == Event.Arg1) { flag = true; break; }
            ++itr;
        }      
        if ( !flag ) Lines.push_back(Event.Arg1);  
        break;
      case EVENT_LINE_DEL:
        for (SimpleList<uint8_t>::iterator itr = Lines.begin(); itr != Lines.end();){
            if ((*itr) == Event.Arg1) { 
                 itr = Lines.erase(itr);
                  flag = true; 
//                  continue;
                  break; 
            }
            ++itr;
        }     
        if (flag) Lines.shrink_to_fit();
        break;             
      default: 
        break;
    }    
    
    #ifdef DRAW_ON
    

    
    u8g.firstPage();  
    do {
    #endif
           u8g.setFontRefHeightExtendedText();
           u8g.setDefaultForegroundColor();
           u8g.setFontPosTop();  
           drawHeader();
           drawTime(CurrentHours, CurrentMinutes);
           drawScLines();

    
    #ifdef DRAW_ON
    } while( u8g.nextPage() );
    #endif

  }
}

void EEPROMSaveCV(int CV, int Value) {
  lncv[CV] = Value;
  Serial.print("EEPROMSaveCV cv:");
  Serial.print(CV);
  Serial.print(" val:");
  Serial.println(Value);
  EEPROM.writeBlock(sizeof(EEPROM_KEY) + sizeof(lncv[0]) * CV, lncv[CV]);
}

void commitLNCVUpdate() {
  Serial.print("Module Address is now: ");
  Serial.println(lncv[CV_ADDR]);
}

void notifyFastClock( uint8_t Rate, uint8_t Day, uint8_t Hour, uint8_t Minute, uint8_t Sync ){
  Serial.print("FastClock H:");
  Serial.print(Hour);
  Serial.print(", Min:");
  Serial.println(Minute);
  if ((Hour != CurrentHours) && (Minute != CurrentMinutes)) {
//      CurrentHours = Hour;
//      CurrentMinutes = Minute;
      addScheduleEvent(EVENT_TIME_UPDATE, 0, 0);    
    }
}

void setup() {
  // Configure the serial port for 57600 baud
  Serial.begin(57600);
  Serial.print("LocoNet Schedule 64x32 LED display with I2C. Shalex. 2017");
  Serial.println(EEPROM_VERSION);

  // First initialize the LocoNet interface
  LocoNet.init(LOCONET_TX_PIN);
  // Initialize the Fast Clock
//  FastClock.init(0, 0, 1);
  // Poll the Current Time from the Command Station
 // FastClock.poll();
  

unsigned long key;

  EEPROM.readBlock(0, key);

  if (key != EEPROM_KEY) {
    
    EEPROM.writeBlock(0,EEPROM_KEY);
    EEPROM.writeBlock(sizeof(EEPROM_KEY),lncv);
    Serial.print("INIT ADDR:");
    EEPROM.writeBlock(sizeof(EEPROM_KEY) + sizeof(lncv),lncv);
  }
  else 
    EEPROM.readBlock(sizeof(EEPROM_KEY),lncv);
  commitLNCVUpdate();
  programmingMode = false;
  
  Lines.reserve(MAX_LINES); 
  CurrentHours = 12;
  CurrentMinutes = 46;
  #ifdef DRAW_ON
  u8g.setContrast(lncv[CV_CONTRAST]);
  #endif
  addScheduleEvent(EVENT_TIME_UPDATE, 0, 0);    
  pinMode(ALARM_LED, OUTPUT);
  digitalWrite(ALARM_LED, LOW);
  

  
}


void loop() {  
  // Check for any received LocoNet packets
  LnPacket = LocoNet.receive();
  if (LnPacket) {
  //  FastClock.processMessage(LnPacket);
    uint8_t packetConsumed(LocoNet.processSwitchSensorMessage(LnPacket));
    if (packetConsumed == 0) {
      Serial.print("Loop ");
      Serial.print((int)LnPacket);
      dumpPacket(LnPacket->ub);
      
      packetConsumed = lnCV.processLNCVMessage(LnPacket);
      
      Serial.print("End Loop\n");
    }
  }
/*  
  if(isTime(&LastFastClockTick, 67)) {
      FastClock.process66msActions(); 
  }
  */
// programmode indication
  ProgrammingModeIdle();
//schedule redraw  
  PictureIdle();  

  TestIdle();
}

void ProgrammingModeIdle(){
unsigned long CurrentMillis;
  
  if (programmingMode){
     CurrentMillis = millis();
     if (CurrentMillis - LedLastBlink > ledCycle) {
        LedState = !LedState;
        LedLastBlink = CurrentMillis;
        digitalWrite(ALARM_LED, LedState);
     }
     if (CurrentMillis - programmingModeStart > PROG_MODE_TIMEOUT) {
        LedState = LOW;
        digitalWrite(ALARM_LED, LedState);
        programmingMode = false;
        programmingModeStart = 0;
        Serial.print("End Programing Mode (Auto).\n");
     }
  }
}


  // This call-back function is called from LocoNet.processSwitchSensorMessage
  // for all Switch Request messages
void notifySwitchRequest( uint16_t Address, uint8_t Output, uint8_t Direction ) {
  Serial.print("Switch Request: ");
  Serial.print(Address, DEC);
  Serial.print(':');
  Serial.print(Direction ? "Closed" : "Thrown");
  Serial.print(" - ");
  Serial.println(Output ? "On" : "Off");
  if ((Address >= lncv[CV_ADDR]) && (Address <= (lncv[CV_ADDR] + MAX_LINES) )) {
    if (Direction) addScheduleEvent(EVENT_LINE_ADD, Address - lncv[CV_ADDR] , 0);
      else addScheduleEvent(EVENT_LINE_DEL, Address - lncv[CV_ADDR] , 0);
  }
}

 
  /**
   * Notifies the code on the reception of a read request.
   * Note that without application knowledge (i.e., art.-nr., module address
   * and "Programming Mode" state), it is not possible to distinguish
   * a read request from a programming start request message.
   */
int8_t notifyLNCVread(uint16_t ArtNr, uint16_t lncvAddress, uint16_t,
    uint16_t & lncvValue) {
  Serial.print("Enter notifyLNCVread(");
  Serial.print(ArtNr, HEX);
  Serial.print(", ");
  Serial.print(lncvAddress, HEX);
  Serial.print(", ");
  Serial.print(lncvValue, HEX);
 Serial.print(")");
  // Step 1: Can this be addressed to me?
  // All ReadRequests contain the ARTNR. For starting programming, we do not accept the broadcast address.
  if (programmingMode) {
    if (ArtNr == ARTNR) {
      if (lncvAddress < LNCV_COUNT) {
        lncvValue = lncv[lncvAddress];
        Serial.print(" LNCV Value: ");
        Serial.print(lncvValue);
        Serial.print("\n");
        return LNCV_LACK_OK;
      } else {
        // Invalid LNCV address, request a NAXK
        return LNCV_LACK_ERROR_UNSUPPORTED;
      }
    } else {
      Serial.print("ArtNr invalid.\n");
      return -1;
    }
  } else {
    Serial.print("Ignoring Request.\n");
    return -1;
  }
}

int8_t notifyLNCVprogrammingStart(uint16_t & ArtNr, uint16_t & ModuleAddress) {
  // Enter programming mode. If we already are in programming mode,
  // we simply send a response and nothing else happens.
  Serial.print("notifyLNCVProgrammingStart artn:");
  Serial.print(ArtNr);
  if (ArtNr == ARTNR) {
    Serial.print("artnrOK ");
    if (ModuleAddress == lncv[0]) {
      Serial.print("moduleUNI ENTERING PROGRAMMING MODE\n");
      programmingMode = true;
      programmingModeStart = millis();
      return LNCV_LACK_OK;
    } else if (ModuleAddress == 0xFFFF) {
      Serial.print("moduleBC ENTERING PROGRAMMING MODE\n");
      ModuleAddress = lncv[0];
      return LNCV_LACK_OK;
    }
  }
  else 
  Serial.print("Ignoring Request.\n");
  return -1;
}

  /**
   * Notifies the code on the reception of a write request
   */
int8_t notifyLNCVwrite(uint16_t ArtNr, uint16_t lncvAddress,
    uint16_t lncvValue) {
  Serial.print("notifyLNCVwrite, ");
  //  dumpPacket(ub);
  if (!programmingMode) {
    Serial.print("not in Programming Mode.\n");
    return -1;
  }

  if (ArtNr == ARTNR) {
    Serial.print("Artnr OK, ");

    if (lncvAddress < LNCV_COUNT) {
      lncv[lncvAddress] = lncvValue;
      programmingModeStart = millis();
      EEPROMSaveCV(lncvAddress, lncvValue);
      return LNCV_LACK_OK;
    }
    else {
      return LNCV_LACK_ERROR_UNSUPPORTED;
    }

  }
  else {
    Serial.print("Artnr Invalid.\n");
    return -1;
  }
}

  /**
   * Notifies the code on the reception of a request to end programming mode
   */
void notifyLNCVprogrammingStop(uint16_t ArtNr, uint16_t ModuleAddress) {
  Serial.print("notifyLNCVprogrammingStop ");
  if (programmingMode) {
    if (ArtNr == ARTNR && ModuleAddress == lncv[0]) {
      programmingMode = false;
      programmingModeStart = 0;
      Serial.print("End Programing Mode.\n");
      commitLNCVUpdate();
    }
    else {
      if (ArtNr != ARTNR) {
        Serial.print("Wrong Artnr.\n");
        return;
      }
      if (ModuleAddress != lncv[0]) {
        Serial.print("Wrong Module Address.\n");
        return;
      }
    }
  }
  else {
    Serial.print("Ignoring Request.\n");
  }
}


void dumpPacket(UhlenbrockMsg & ub) {
  Serial.print(" PKT: ");
  Serial.print(ub.command, HEX);
  Serial.print(" ");
  Serial.print(ub.mesg_size, HEX);
  Serial.print(" ");
  Serial.print(ub.SRC, HEX);
  Serial.print(" ");
  Serial.print(ub.DSTL, HEX);
  Serial.print(" ");
  Serial.print(ub.DSTH, HEX);
  Serial.print(" ");
  Serial.print(ub.ReqId, HEX);
  Serial.print(" ");
  Serial.print(ub.PXCT1, HEX);
  Serial.print(" ");
  for (int i(0); i < 8; ++i) {
//     Serial.print(ub.Data[i], HEX);
    Serial.print(" ");
  }
  Serial.write("\n");
}
